from time import time



N = 5000; D = 3 ; # Cloud of 10,000 points in 3D
import numpy as np

q = np.random.rand(N,D)
p = np.random.rand(N,D)

t = time()

# Actual computations.
q_i  = q[:,np.newaxis,:] # shape (N,D) -> (N,1,D)
q_j  = q[np.newaxis,:,:] # shape (N,D) -> (1,N,D)

sqd  = np.sum( (q_i - q_j)**2 , 2 ) # |q_i-q_j|^2
K_qq = np.exp( - sqd  ) # Gaussian kernel of dev. 1
v    = K_qq @ p # matrix mult. (N,N)@(N,D) = (N,D)
#
# Finally, compute the Hamiltonian H(q,p): .5*<p,v>
H    = .5 * np.inner( p.ravel(), v.ravel() )

print("H =", H)
print("Elapsed time : ", time() - t, 's')



import torch         # GPU + autodiff library
from   torch.autograd import Variable, grad

# With PyTorch, using the GPU is that simple:
use_gpu  = torch.cuda.is_available()
dtype    = torch.cuda.FloatTensor if use_gpu \
           else torch.FloatTensor

# Generate arbitrary arrays on the CPU or GPU:
q = torch.from_numpy( q ).type(dtype)
p = torch.from_numpy( p ).type(dtype)

# Wrap them into "autodiff" graph nodes.
q = Variable( q, requires_grad = True )
p = Variable( p, requires_grad = True )

t = time()

# Actual computations.
q_i  = q.unsqueeze(1) # shape (N,D) -> (N,1,D)
q_j  = q.unsqueeze(0) # shape (N,D) -> (1,N,D)

sqd  = torch.sum( (q_i - q_j)**2 , 2 ) # |q_i-q_j|^2
K_qq = torch.exp( - sqd )              # Gaussian kernel
v    = K_qq @ p # matrix mult. (N,N)@(N,D) = (N,D)
#
# Finally, compute the Hamiltonian H(q,p): .5*<p,v>
H    = .5 * torch.dot( p.view(-1), v.view(-1) )

print("H =", H.data.cpu().numpy()[0])
print("Elapsed time : ", time() - t, 's')
#
# Automatic differentiation is straitghtforward

t = time()

[dq,dp] = grad( H, [q,p])
print("dq =", dq.data.cpu().numpy())
print("Elapsed time : ", time() - t, 's')
print(time(), t)
